@@ -1,12 +1,10 @@
-import f4
-import fastnumbers
-import glob
-from itertools import chain
-from joblib import Parallel, delayed
-import math
-import os
-import sys
-import zstandard
+# import f4
+# import fastnumbers
+# import glob
+# from itertools import chain
+# from joblib import Parallel, delayed
+# import math
+# import sys
 
 class Parser:
     """
@@ -18,37 +16,44 @@ class Parser:
     Attributes:
         data_file_path (str): The path to an existing F4 file.
     """
-    #def __init__(self, data_file_path, fixed_file_extensions=["", ".cc", ".ct"], stats_file_extensions=[".ll", ".mccl", ".nrow", ".ncol"]):
-    def __init__(self, data_file_path, fixed_file_extensions=["", ".cc", ".ct"], stats_file_extensions=[".ll", ".mccl"]):
-        #TODO: expand this out for the other parameters.
+    def __init__(self, data_file_path):
         if not isinstance(data_file_path, str):
             raise Exception("You must specify data_file_path as an str value.")
 
         self.data_file_path = data_file_path
+        self.__file_handle = f4.open_read_file(self.data_file_path)
 
-        # Cache file handles in a dictionary.
-        self.__file_handles = {}
-        for ext in fixed_file_extensions:
-            self.__file_handles[ext] = self._set_file_handle(ext)
+        file_map_length_string = self.__file_handle.readline()
+        file_map_length = fastnumbers.fast_int(file_map_length_string.rstrip(b"\n"))
+        self.__file_map_dict = f4.deserialize(self.__file_handle[len(file_map_length_string):(len(file_map_length_string) + file_map_length)])
+
+        # Invert the abbreviations used for file keys in the dictionary so they are more human readable.
+        for file_name, abbreviation in f4.FILE_KEY_ABBREVIATIONS.items():
+            if abbreviation in self.__file_map_dict:
+                self.__file_map_dict[file_name] = self.__file_map_dict[abbreviation]
+                del self.__file_map_dict[abbreviation]
 
         # Cache statistics in a dictionary.
         self.__stats = {}
-        for ext in stats_file_extensions:
-            self.__stats[ext] = f4.read_int_from_file(data_file_path, ext)
+
+        data_size =  self.__file_map_dict["data"][1] - self.__file_map_dict["data"][0]
+        self.__stats["nrow"] = fastnumbers.fast_int(data_size / self._get_stat("ll"))
+
+        cc_size = self.__file_map_dict["cc"][1] - self.__file_map_dict["cc"][0]
+        self.__stats["ncol"] = fastnumbers.fast_int(cc_size / self._get_stat("mccl")) - 1
 
     def __enter__(self):
         return self
 
     def __exit__(self, the_type, value, traceback):
-        for handle in self.__file_handles.values():
-            handle.close()
+        self.__file_handle.close()
 
-    def query_and_save(self, fltr, select_columns, out_file_path=None, out_file_type="tsv", num_processes=1, lines_per_chunk=10):
+    def query_and_write(self, fltr, select_columns, out_file_path=None, out_file_type="tsv", num_processes=1, lines_per_chunk=10):
         """
         Query the data file using zero or more filters.
 
         This function accepts filtering criteria, identifies matching rows,
-        and saves the output (for select columns) to an output file.
+        and writes the output (for select columns) to an output file or standard output.
 
         Args:
             fltr (BaseFilter): A filter.
@@ -116,10 +121,10 @@ class Parser:
             # Write output (in chunks)
             with open(out_file_path, 'wb') as out_file:
                 out_file.write(b"\t".join(select_columns) + b"\n") # Header line
-                
+
                 out_lines = []
                 for row_index in keep_row_indices:
-                    out_values = parse_function(row_index, select_column_coords, decompression_type, decompressor, bigram_size_dict, select_columns)
+                    out_values = parse_function(row_index, select_column_coords, decompression_type=decompression_type, decompressor=decompressor, bigram_size_dict=bigram_size_dict, column_names=select_columns)
                     out_lines.append(b"\t".join(out_values))
 
                     if len(out_lines) % lines_per_chunk == 0:
@@ -132,7 +137,7 @@ class Parser:
             sys.stdout.buffer.write(b"\t".join(select_columns) + b"\n") # Header line
 
             for row_index in keep_row_indices:
-                out_values = parse_function(row_index, select_column_coords, decompression_type, decompressor, bigram_size_dict, select_columns)
+                out_values = parse_function(row_index, select_column_coords, decompression_type=decompression_type, decompressor=decompressor, bigram_size_dict=bigram_size_dict, column_names=select_columns)
                 sys.stdout.buffer.write(b"\t".join(out_values))
 
                 if row_index != keep_row_indices[-1]:
@@ -141,96 +146,86 @@ class Parser:
     def head(self, n = 10, select_columns=None, out_file_path=None, out_file_type="tsv"):
         if not select_columns:
             select_columns = []
-        self.query_and_save(f4.HeadFilter(n, select_columns), select_columns, out_file_path=out_file_path, out_file_type=out_file_type)
+        self.query_and_write(f4.HeadFilter(n, select_columns), select_columns, out_file_path=out_file_path, out_file_type=out_file_type)
 
     def tail(self, n = 10, select_columns=None, out_file_path=None, out_file_type="tsv"):
         if not select_columns:
             select_columns = []
-        self.query_and_save(f4.TailFilter(n, select_columns), select_columns, out_file_path=out_file_path, out_file_type=out_file_type)
+        self.query_and_write(f4.TailFilter(n, select_columns), select_columns, out_file_path=out_file_path, out_file_type=out_file_type)
 
     def get_num_rows(self):
-        return int(len(self.__file_handles[""]) / self.__stats[".ll"])
-        #return self.__stats[".nrow"]
+        return self._get_stat("nrow")
 
     def get_num_cols(self):
-        return int(len(self.__file_handles[".cc"]) / self.__stats[".mccl"]) - 1
-        #return self.__stats[".ncol"]
+        return self._get_stat("ncol")
 
     def get_column_type_from_name(self, column_name):
         try:
-            with f4.IndexSearcher._get_index_parser(f"{self.data_file_path}.cn") as index_parser:
-                return self._get_column_type_from_index(self._get_column_index_from_name(index_parser, column_name))
+            return self._get_column_type_from_index(self._get_column_index_from_name(column_name))
         except:
             raise Exception(f"A column with the name {column_name} does not exist.")
 
-    def _get_column_index_from_name(self, index_parser, column_name):
-        position = f4.IndexSearcher._get_identifier_row_index(index_parser, column_name.encode(), self.get_num_cols())
+    ##############################################
+    # Non-public functions
+    ##############################################
+
+    def _get_column_index_from_name(self, column_name):
+        position = f4.IndexSearcher._get_identifier_row_index(self, column_name.encode(), self.get_num_cols(), data_prefix="cn")
 
         if position < 0:
             raise Exception(f"Could not retrieve index because column named {column_name} was not found.")
 
         return position
 
-    def _get_file_handle(self, ext):
-        return self.__file_handles[ext]
+    def _get_stat(self, stat):
+        if stat not in self.__stats:
+            self.__stats[stat] = fastnumbers.fast_int(self._get_stat_from_file(stat))
 
-    def _set_file_handle(self, ext):
-        if ext not in self.__file_handles:
-            self.__file_handles[ext] = f4.open_read_file(self.data_file_path, ext)
+        return self.__stats[stat]
 
-        return self._get_file_handle(ext)
-
-    def _get_stat(self, ext):
-        return self.__stats[ext]
-
-    ##############################################
-    # Non-public functions
-    ##############################################
+    def _get_stat_from_file(self, key):
+        coordinates = self.__file_map_dict[key]
+        return self.__file_handle[coordinates[0]:coordinates[1]]
 
     def _get_column_type_from_index(self, column_index):
-        #return next(self.__parse_data_values(column_index, 2, [[0, 1]], self.__file_handles[".ct"])).decode()
-        return next(self._parse_data_values(column_index, 1, [[0, 1]], self.__file_handles[".ct"])).decode()
+        return next(self._parse_data_values_from_file(column_index, 1, [[0, 1]], "", "ct")).decode()
 
     def _get_column_meta(self, filter_column_set, select_columns):
         column_type_dict = {}
         column_coords_dict = {}
         column_index_name_dict = {}
-        #column_name_index_dict = {} #TODO
 
         if len(select_columns) == 0:
-            with f4.Parser(self.data_file_path + ".cn", fixed_file_extensions=["", ".cc"], stats_file_extensions=[".ll", ".mccl"]) as cn_parser:
-                coords = cn_parser._parse_data_coords([0, 1])
+            coords = self._parse_data_coords([0, 1], data_prefix="cn")
 
-                for row_index in range(self.get_num_cols()):
-                    values = cn_parser._parse_row_values(row_index, coords)
-                    column_name = values[0]
-                    column_index = fastnumbers.fast_int(values[1])
+            for row_index in range(self.get_num_cols()):
+                values = self._parse_row_values(row_index, coords, data_prefix="cn")
+                column_name = values[0]
+                column_index = fastnumbers.fast_int(values[1])
 
-                    column_index_name_dict[column_index] = column_name
-                    #column_name_index_dict[column_name] = column_index
+                column_index_name_dict[column_index] = column_name
 
-                    if column_name in filter_column_set:
-                        column_type_dict[column_name] = row_index
+                if column_name in filter_column_set:
+                    column_type_dict[column_name] = row_index
 
-                all_coords = self._parse_data_coords(range(self.get_num_cols()))
-                for row_index in range(self.get_num_cols()):
-                    column_coords_dict[column_index_name_dict[row_index]] = all_coords[row_index]
+            all_coords = self._parse_data_coords(range(self.get_num_cols()))
+            for row_index in range(self.get_num_cols()):
+                column_coords_dict[column_index_name_dict[row_index]] = all_coords[row_index]
 
             all_columns = [x[1] for x in sorted(column_index_name_dict.items())]
             select_columns = all_columns
         else:
-            with f4.IndexSearcher._get_index_parser(f"{self.data_file_path}.cn") as index_parser:
-                select_columns = [name.encode() for name in select_columns]
-                all_columns = list(filter_column_set | set(select_columns))
+            select_columns = [name.encode() for name in select_columns]
+            all_columns = list(filter_column_set | set(select_columns))
 
-                column_name_index_dict = {}
-                for column_name in all_columns:
-                    column_index = self._get_column_index_from_name(index_parser, column_name.decode())
-                    column_name_index_dict[column_name] = column_index
-                    column_index_name_dict[column_index] = column_name
+            column_name_index_dict = {}
+            for column_name in all_columns:
+                column_index = self._get_column_index_from_name(column_name.decode())
+                column_name_index_dict[column_name] = column_index
+                column_index_name_dict[column_index] = column_name
 
-                for column_name in filter_column_set:
-                    column_type_dict[column_name] = self._get_column_type_from_index(column_name_index_dict[column_name])
+            for column_name in filter_column_set:
+                column_type_dict[column_name] = self._get_column_type_from_index(column_name_index_dict[column_name])
 
             all_column_indices = [column_name_index_dict[name] for name in all_columns]
             all_coords = self._parse_data_coords(all_column_indices)
@@ -241,19 +236,16 @@ class Parser:
         decompression_type = None
         decompressor = None
         bigram_size_dict = {}
-        decompressor_file_path = f"{self.data_file_path}.cmpr"
 
-        if os.path.exists(decompressor_file_path):
-            decompression_text = f4.read_str_from_file(decompressor_file_path)
+        if "cmpr" in self.__file_map_dict:
+            decompression_text_coords = self.__file_map_dict["cmpr"]
+            decompression_text = self.__file_handle[decompression_text_coords[0]:decompression_text_coords[1]]
 
             if decompression_text == b"z":
                 decompression_type = "zstd"
-                # decompressor = zstandard.ZstdDecompressor()
             else:
                 decompression_type = "dictionary"
-                decompressor = self._get_decompression_dict(decompressor_file_path, column_index_name_dict)
-                # if len(decompression_dict) > 0:
-                #    select_compression_dict = self._invert_decompression_dict(decompression_dict, select_columns)
+                decompressor = f4.deserialize(decompression_text)
 
                 for column_name in all_columns:
                     bigram_size_dict[column_name] = f4.get_bigram_size(len(decompressor[column_name]["map"]))
@@ -275,11 +267,12 @@ class Parser:
         if len(row_indices) > 0:
             yield row_indices
 
-    def _parse_data_coords(self, indices):
+    def _parse_data_coords(self, indices, data_prefix=""):
+        #TODO: Store coordinates in an object-level dictionary after retrieving them?
+        file_key = "cc"
         data_coords = []
         out_dict = {}
-        #mccl = self.__stats[".mccl"] + 1
-        mccl = self.__stats[".mccl"]
+        mccl = self._get_stat(data_prefix + "mccl")
 
         for index in indices:
             start_pos = index * mccl
@@ -291,7 +284,7 @@ class Parser:
                 data_start_pos = out_dict[index]
             # If not, retrieve the start position from the cc file and then cache it.
             else:
-                data_start_pos = fastnumbers.fast_int(self.__file_handles[".cc"][start_pos:next_start_pos].rstrip(b" "))
+                data_start_pos = fastnumbers.fast_int(self._parse_data_value_from_file(0, 0, [start_pos, next_start_pos], data_prefix, file_key).rstrip(b" "))
                 out_dict[index] = data_start_pos
 
             # See if we already have cached the end position.
@@ -299,23 +292,26 @@ class Parser:
                 data_end_pos = out_dict[index + 1]
             # If not, retrieve the end position from the cc file and then cache it.
             else:
-                data_end_pos = fastnumbers.fast_int(self.__file_handles[".cc"][next_start_pos:further_next_start_pos].rstrip(b" "))
+                data_end_pos = fastnumbers.fast_int(self._parse_data_value_from_file(0, 0, [next_start_pos, further_next_start_pos], data_prefix, file_key).rstrip(b" "))
                 out_dict[index + 1] = data_end_pos
 
             data_coords.append([data_start_pos, data_end_pos])
 
         return data_coords
 
-    def _parse_data_value(self, start_element, segment_length, coords, str_like_object):
-        start_pos = start_element * segment_length
+    def _parse_data_value_from_file(self, start_element, segment_length, coords, data_prefix, file_key):
+        start_pos = start_element * segment_length + self.__file_map_dict[data_prefix + file_key][0]
+        return self.__file_handle[(start_pos + coords[0]):(start_pos + coords[1])]
 
-        return str_like_object[(start_pos + coords[0]):(start_pos + coords[1])]
+    def _parse_data_values_from_string(self, data_coords, string):
+        for coords in data_coords:
+            yield string[(coords[0]):(coords[1])].rstrip(b" ")
 
-    def _parse_data_values(self, start_element, segment_length, data_coords, str_like_object):
-        start_pos = start_element * segment_length
+    def _parse_data_values_from_file(self, start_element, segment_length, data_coords, data_prefix, file_key):
+        start_pos = start_element * segment_length + self.__file_map_dict[data_prefix + file_key][0]
 
         for coords in data_coords:
-            yield str_like_object[(start_pos + coords[0]):(start_pos + coords[1])].rstrip(b" ")
+            yield self.__file_handle[(start_pos + coords[0]):(start_pos + coords[1])].rstrip(b" ")
 
     def _get_parse_row_value_function(self, decompression_type):
         if not decompression_type:
@@ -325,39 +321,17 @@ class Parser:
         else:
             return self._parse_dictionary_compressed_row_value
 
-    def _parse_row_value(self, row_index, column_coords, line_length, file_handle, decompression_type=None, decompressor=None, bigram_size_dict=None, column_name=None):
-        return self._parse_data_value(row_index, line_length, column_coords, file_handle).rstrip(b" ")
+    def _parse_row_value(self, row_index, column_coords, line_length, data_prefix="", decompression_type=None, decompressor=None, bigram_size_dict=None, column_name=None):
+        return self._parse_data_value_from_file(row_index, line_length, column_coords, data_prefix, "data").rstrip(b" ")
 
-        # if decompression_type == "zstd":
-        #     line = self.__parse_data_value(row_index, line_length, [0, line_length], file_handle)
-        #     line = decompressor.decompress(line)
-        #     return self.__parse_data_value(0, 0, column_coords, line).rstrip(b" ")
-        # else:
-        #     value = self.__parse_data_value(row_index, line_length, column_coords, file_handle).rstrip(b" ")
-        #
-        #     if decompression_type == "dictionary":
-        #         value = f4.decompress(value, decompressor[column_name], bigram_size_dict[column_name])
-        #
-        #     return value
-
-    def _parse_zstd_compressed_row_value(self, row_index, column_coords, line_length, file_handle, decompression_type=None, decompressor=None, bigram_size_dict=None, column_name=None):
-        # if decompression_type == "zstd":
-        #     line = self.__parse_data_value(row_index, line_length, [0, line_length], file_handle)
-        #     line = decompressor.decompress(line)
-        #     return self.__parse_data_value(0, 0, column_coords, line).rstrip(b" ")
-        # else:
-        #     value = self.__parse_data_value(row_index, line_length, column_coords, file_handle).rstrip(b" ")
-        #
-        #     if decompression_type == "dictionary":
-        #         value = f4.decompress(value, decompressor[column_name], bigram_size_dict[column_name])
-        #
-        #     return value
-        line = self._parse_data_value(row_index, line_length, [0, line_length], file_handle)
+    def _parse_zstd_compressed_row_value(self, row_index, column_coords, line_length, data_prefix="", decompression_type=None, decompressor=None, bigram_size_dict=None, column_name=None):
+        line = self._parse_data_value_from_file(row_index, line_length, [0, line_length], data_prefix, "data")
         line = decompressor.decompress(line)
-        return self._parse_data_value(0, 0, column_coords, line).rstrip(b" ")
 
-    def _parse_dictionary_compressed_row_value(self, row_index, column_coords, line_length, file_handle, decompression_type=None, decompressor=None, bigram_size_dict=None, column_name=None):
-        value = self._parse_data_value(row_index, line_length, column_coords, file_handle).rstrip(b" ")
+        return line[column_coords[0]:column_coords[1]].rstrip(b" ")
+
+    def _parse_dictionary_compressed_row_value(self, row_index, column_coords, line_length, data_prefix="", decompression_type=None, decompressor=None, bigram_size_dict=None, column_name=None):
+        value = self._parse_data_value_from_file(row_index, line_length, column_coords, data_prefix, "data").rstrip(b" ")
         return f4.decompress(value, decompressor[column_name], bigram_size_dict[column_name])
 
     def _get_parse_row_values_function(self, decompression_type):
@@ -368,23 +342,24 @@ class Parser:
         else:
             return self._parse_dictionary_compressed_row_values
 
-    def _parse_row_values(self, row_index, column_coords, decompression_type=None, decompressor=None, bigram_size_dict=None, column_names=None):
-        return list(self._parse_data_values(row_index, self.__stats[".ll"], column_coords, self.__file_handles[""]))
+    def _parse_row_values(self, row_index, column_coords, data_prefix="", decompression_type=None, decompressor=None, bigram_size_dict=None, column_names=None):
+        return list(self._parse_data_values_from_file(row_index, self._get_stat(data_prefix + "ll"), column_coords, data_prefix, "data"))
 
-    def _parse_zstd_compressed_row_values(self, row_index, column_coords, decompression_type=None, decompressor=None, bigram_size_dict=None, column_names=None):
-        line_length = self.__stats[".ll"]
-        line = self._parse_data_value(row_index, line_length, [0, line_length], self.__file_handles[""])
+    def _parse_zstd_compressed_row_values(self, row_index, column_coords, data_prefix="", decompression_type=None, decompressor=None, bigram_size_dict=None, column_names=None):
+        line_length = self._get_stat("ll")
+        line = self._parse_data_value_from_file(row_index, line_length, [0, line_length], data_prefix, "data")
         line = decompressor.decompress(line)
 
-        return list(self._parse_data_values(0, 0, column_coords, line))
+        return list(self._parse_data_values_from_string(column_coords, line))
+
+    def _parse_dictionary_compressed_row_values(self, row_index, column_coords, data_prefix="", decompression_type=None, decompressor=None, bigram_size_dict=None, column_names=None):
+            values = list(self._parse_data_values_from_file(row_index, self._get_stat("ll"), column_coords, data_prefix, "data"))
 
-    def _parse_dictionary_compressed_row_values(self, row_index, column_coords, decompression_type=None, decompressor=None, bigram_size_dict=None, column_names=None):
-            values = list(self._parse_data_values(row_index, self.__stats[".ll"], column_coords, self.__file_handles[""]))
             return [f4.decompress(values.pop(0), decompressor[column_name], bigram_size_dict[column_name]) for column_name in column_names]
 
-    def _get_decompression_dict(self, file_path, column_index_name_dict):
-        with open(file_path, "rb") as cmpr_file:
-            return f4.deserialize(cmpr_file.read())
+    # def _get_decompression_dict(self, file_path, column_index_name_dict):
+    #     with open(file_path, "rb") as cmpr_file:
+    #         return f4.deserialize(cmpr_file.read())
 
     #     compression_dict = {}
     #     with open(file_path, "rb") as cmpr_file:
@@ -400,14 +375,14 @@ class Parser:
     #     #     compression_dict[column_index_name_dict[column_index]] = {}
     #     #     compression_dict[column_index_name_dict[column_index]]["map"] = {}
 
-    #     # with Parser(file_path, fixed_file_extensions=["", ".cc"], stats_file_extensions=[".ll", ".mccl"]) as parser:
+    #     # with Parser(file_path, fixed_file_extensions=[".data", ".cc"], stats_file_extensions=[".ll", ".mccl"]) as parser:
     #     #     coords = parser._parse_data_coords([0, 1, 2])
-    #     #     num_rows = fastnumbers.fast_int((len(parser.get_file_handle("")) + 1) / parser.get_stat(".ll"))
+    #     #     num_rows = fastnumbers.fast_int((len(parser.get_file_handle(".data")) + 1) / parser.get_stat(".ll"))
 
     #     #     # Use a set for performance reasons
     #     #     column_indices_set = set(column_index_name_dict.keys())
 
-    #     #     with Parser(f"{self.data_file_path}.cmpr", fixed_file_extensions=["", ".cc"], stats_file_extensions=[".ll", ".mccl"]) as parser:
+    #     #     with Parser(f"{self.data_file_path}.cmpr", fixed_file_extensions=[".data", ".cc"], stats_file_extensions=[".ll", ".mccl"]) as parser:
     #     #         for row_index in range(num_rows):
     #     #             values = parser.__parse_row_values(row_index, coords)
     #     #             column_index = fastnumbers.fast_int(values[0])
@@ -422,7 +397,7 @@ class Parser:
     #     # # for i, column_name in enumerate(column_names):
     #     # #     compression_dict2[column_name] = compression_dict[column_indices[i]]
 
-    #     # with Parser(f"{self.data_file_path}.cmprtype", fixed_file_extensions=[""], stats_file_extensions=[".ll"]) as parser:
+    #     # with Parser(f"{self.data_file_path}.cmprtype", fixed_file_extensions=[".data"], stats_file_extensions=[".ll"]) as parser:
     #     #     coords = [[0, 1]]
 
     #     #     for column_index in column_index_name_dict.keys():
